""" opentrons.protocol_api.labware: classes and functions for labware handling

This module provides things like :py:class:`Labware`, and :py:class:`Well`
to encapsulate labware instances used in protocols
and their wells. It also provides helper functions to load and save labware
and labware calibration offsets. It contains all the code necessary to
transform from labware symbolic points (such as "well a1 of an opentrons
tiprack") to points in deck coordinates.
"""
import logging

from pathlib import Path
from itertools import dropwhile
from typing import Any, AnyStr, List, Dict, Optional, Union, Tuple, TYPE_CHECKING


from opentrons.protocols.api_support.util import requires_version, labware_column_shift
from opentrons.protocols.context.labware import AbstractLabware
from opentrons.protocols.geometry.well_geometry import WellGeometry
from opentrons.protocols import labware as labware_module
from opentrons.protocols.context.well import WellImplementation
from opentrons.types import Location, Point, LocationLabware
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.geometry.deck_item import DeckItem

if TYPE_CHECKING:
    from opentrons.protocols.geometry.module_geometry import (  # noqa: F401
        ModuleGeometry,
    )
    from opentrons_shared_data.labware.dev_types import (
        LabwareDefinition,
        LabwareParameters,
    )


MODULE_LOG = logging.getLogger(__name__)


class TipSelectionError(Exception):
    pass


class OutOfTipsError(Exception):
    pass


class Well:
    """
    The Well class represents a  single well in a :py:class:`Labware`

    It provides functions to return positions used in operations on the well
    such as :py:meth:`top`, :py:meth:`bottom`
    """

    def __init__(
        self,
        well_implementation: WellImplementation,
        api_level: Optional[APIVersion] = None,
    ):
        """
        Create a well, and track the Point corresponding to the top-center of
        the well (this Point is in absolute deck coordinates)

        """
        self._api_version = api_level or MAX_SUPPORTED_VERSION
        self._impl = well_implementation
        self._geometry = well_implementation.get_geometry()

    @property  # type: ignore
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        return self._api_version

    @property  # type: ignore
    @requires_version(2, 0)
    def parent(self) -> "Labware":
        return Labware(implementation=self._geometry.parent, api_level=self.api_version)

    @property  # type: ignore
    @requires_version(2, 0)
    def has_tip(self) -> bool:
        return self._impl.has_tip()

    @has_tip.setter
    def has_tip(self, value: bool):
        self._impl.set_has_tip(value)

    @property
    def max_volume(self) -> float:
        return self._geometry.max_volume

    @property
    def geometry(self) -> WellGeometry:
        return self._geometry

    @property  # type: ignore
    @requires_version(2, 0)
    def diameter(self) -> Optional[float]:
        return self._geometry.diameter

    @property  # type: ignore
    @requires_version(2, 9)
    def length(self) -> Optional[float]:
        """
        The length of a well, if the labware has
        square wells.
        """
        return self._geometry._length

    @property  # type: ignore
    @requires_version(2, 9)
    def width(self) -> Optional[float]:
        """
        The width of a well, if the labware has
        square wells.
        """
        return self._geometry._width

    @property  # type: ignore
    @requires_version(2, 9)
    def depth(self) -> float:
        """
        The depth of a well in a labware.
        """
        return self._geometry._depth

    @property
    def display_name(self):
        return self._impl.get_display_name()

    @property  # type: ignore
    @requires_version(2, 7)
    def well_name(self) -> str:
        return self._impl.get_name()

    @requires_version(2, 0)
    def top(self, z: float = 0.0) -> Location:
        """
        :param z: the z distance in mm
        :return: a Point corresponding to the absolute position of the
                 top-center of the well relative to the deck (with the
                 front-left corner of slot 1 as (0,0,0)). If z is specified,
                 returns a point offset by z mm from top-center
        """
        return Location(self._geometry.top(z), self)

    @requires_version(2, 0)
    def bottom(self, z: float = 0.0) -> Location:
        """
        :param z: the z distance in mm
        :return: a Point corresponding to the absolute position of the
                 bottom-center of the well (with the front-left corner of
                 slot 1 as (0,0,0)). If z is specified, returns a point
                 offset by z mm from bottom-center
        """
        return Location(self._geometry.bottom(z), self)

    @requires_version(2, 0)
    def center(self) -> Location:
        """
        :return: a Point corresponding to the absolute position of the center
                 of the well relative to the deck (with the front-left corner
                 of slot 1 as (0,0,0))
        """
        return Location(self._geometry.center(), self)

    @requires_version(2, 8)
    def from_center_cartesian(self, x: float, y: float, z: float) -> Point:
        """
        Specifies an arbitrary point in deck coordinates based
        on percentages of the radius in each axis. For example, to specify the
        back-right corner of a well at 1/4 of the well depth from the bottom,
        the call would be `from_center_cartesian(1, 1, -0.5)`.

        No checks are performed to ensure that the resulting position will be
        inside of the well.

        :param x: a float in the range [-1.0, 1.0] for a percentage of half of
            the radius/length in the X axis
        :param y: a float in the range [-1.0, 1.0] for a percentage of half of
            the radius/width in the Y axis
        :param z: a float in the range [-1.0, 1.0] for a percentage of half of
            the height above/below the center

        :return: a Point representing the specified location in absolute deck
        coordinates
        """
        return self._geometry.from_center_cartesian(x, y, z)

    def _from_center_cartesian(self, x: float, y: float, z: float) -> Point:
        """
        Private version of from_center_cartesian. Present only for backward
        compatibility.
        """
        MODULE_LOG.warning(
            "This method is deprecated. Please use " "'from_center_cartesian' instead."
        )
        return self.from_center_cartesian(x, y, z)

    def __repr__(self):
        return self._impl.get_display_name()

    def __eq__(self, other: object) -> bool:
        """
        Assuming that equality of wells in this system is having the same
        absolute coordinates for the top.
        """
        if not isinstance(other, Well):
            return NotImplemented
        return self.top().point == other.top().point

    def __hash__(self):
        return hash(self.top().point)


class Labware(DeckItem):
    """
    This class represents a labware, such as a PCR plate, a tube rack,
    reservoir, tip rack, etc. It defines the physical geometry of the labware,
    and provides methods for accessing wells within the labware.

    It is commonly created by calling :py:meth:`ProtocolContext.load_labware`.

    To access a labware's wells, you can use its well accessor methods:
    :py:meth:`wells_by_name`, :py:meth:`wells`, :py:meth:`columns`,
    :py:meth:`rows`, :py:meth:`rows_by_name`, and :py:meth:`columns_by_name`.
    You can also use an instance of a labware as a Python dictionary, accessing
    wells by their names. The following example shows how to use all of these
    methods to access well A1:

    .. code-block :: python

       labware = context.load_labware('corning_96_wellplate_360ul_flat', 1)
       labware['A1']
       labware.wells_by_name()['A1']
       labware.wells()[0]
       labware.rows()[0][0]
       labware.columns()[0][0]
       labware.rows_by_name()['A'][0]
       labware.columns_by_name()[0][0]

    """

    def __init__(
        self, implementation: AbstractLabware, api_level: Optional[APIVersion] = None
    ) -> None:
        """
        :param implementation: The class that implements the public interface
                               of the class.
        :param APIVersion api_level: the API version to set for the instance.
                                     The :py:class:`.Labware` will
                                     conform to this level. If not specified,
                                     defaults to
                                     :py:attr:`.MAX_SUPPORTED_VERSION`.
        """
        if not api_level:
            api_level = MAX_SUPPORTED_VERSION
        if api_level > MAX_SUPPORTED_VERSION:
            raise RuntimeError(
                f"API version {api_level} is not supported by this "
                f"robot software. Please either reduce your requested API "
                f"version or update your robot."
            )
        self._api_version = api_level
        self._implementation = implementation

    @property
    def separate_calibration(self) -> bool:
        return self._implementation.separate_calibration

    @property  # type: ignore
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        return self._api_version

    def __getitem__(self, key: str) -> Well:
        return self.wells_by_name()[key]

    @property  # type: ignore
    @requires_version(2, 0)
    def uri(self) -> str:
        """A string fully identifying the labware.

        :returns: The uri, ``"namespace/loadname/version"``
        """
        return self._implementation.get_uri()

    @property  # type: ignore
    @requires_version(2, 0)
    def parent(self) -> LocationLabware:
        """The parent of this labware. Usually a slot name."""
        return self._implementation.get_geometry().parent.labware.object

    @property  # type: ignore
    @requires_version(2, 0)
    def name(self) -> str:
        """Can either be the canonical name of the labware, which is used to
        load it, or the label of the labware specified by a user."""
        return self._implementation.get_name()

    @name.setter
    def name(self, new_name):
        """Set the labware name"""
        self._implementation.set_name(new_name)

    @property  # type: ignore
    @requires_version(2, 0)
    def load_name(self) -> str:
        """The API load name of the labware definition"""
        return self._implementation.load_name

    @property  # type: ignore
    @requires_version(2, 0)
    def parameters(self) -> "LabwareParameters":
        """Internal properties of a labware including type and quirks"""
        return self._implementation.get_parameters()

    @property  # type: ignore
    @requires_version(2, 0)
    def quirks(self) -> List[str]:
        """Quirks specific to this labware."""
        return self._implementation.get_quirks()

    @property  # type: ignore
    @requires_version(2, 0)
    def magdeck_engage_height(self) -> Optional[float]:
        p = self._implementation.get_parameters()
        if not p["isMagneticModuleCompatible"]:
            return None
        else:
            return p["magneticModuleEngageHeight"]

    def set_calibration(self, delta: Point):
        """
        Called by save calibration in order to update the offset on the object.
        """
        self._implementation.set_calibration(delta)

    @property  # type: ignore
    @requires_version(2, 0)
    def calibrated_offset(self) -> Point:
        return self._implementation.get_calibrated_offset()

    @requires_version(2, 0)
    def well(self, idx) -> Well:
        """Deprecated---use result of `wells` or `wells_by_name`"""
        if isinstance(idx, int):
            res = self._implementation.get_wells()[idx]
        elif isinstance(idx, str):
            res = self._implementation.get_wells_by_name()[idx]
        else:
            res = NotImplemented
        return self._well_from_impl(res)

    @requires_version(2, 0)
    def wells(self, *args) -> List[Well]:
        """
        Accessor function used to generate a list of wells in top -> down,
        left -> right order. This is representative of moving down `rows` and
        across `columns` (e.g. 'A1', 'B1', 'C1'...'A2', 'B2', 'C2')

        With indexing one can treat it as a typical python
        list. To access well A1, for example, simply write: labware.wells()[0]

        Note that this method takes args for backward-compatibility, but use
        of args is deprecated and will be removed in future versions. Args
        can be either strings or integers, but must all be the same type (e.g.:
        `self.wells(1, 4, 8)` or `self.wells('A1', 'B2')`, but
        `self.wells('A1', 4)` is invalid.

        :return: Ordered list of all wells in a labware
        """
        if not args:
            res = self._implementation.get_wells()
        elif isinstance(args[0], int):
            res = [self._implementation.get_wells()[idx] for idx in args]
        elif isinstance(args[0], str):
            by_name = self._implementation.get_wells_by_name()
            res = [by_name[idx] for idx in args]
        else:
            raise TypeError
        return [self._well_from_impl(w) for w in res]

    @requires_version(2, 0)
    def wells_by_name(self) -> Dict[str, Well]:
        """
        Accessor function used to create a look-up table of Wells by name.

        With indexing one can treat it as a typical python
        dictionary whose keys are well names. To access well A1, for example,
        simply write: labware.wells_by_name()['A1']

        :return: Dictionary of well objects keyed by well name
        """
        wells = self._implementation.get_wells_by_name()
        return {k: self._well_from_impl(v) for k, v in wells.items()}

    @requires_version(2, 0)
    def wells_by_index(self) -> Dict[str, Well]:
        """
        .. deprecated:: 2.0
            Use :py:meth:`wells_by_name` or dict access instead.
        """
        MODULE_LOG.warning(
            "wells_by_index is deprecated. Use wells_by_name or dict access instead."
        )
        return self.wells_by_name()

    @requires_version(2, 0)
    def rows(self, *args) -> List[List[Well]]:
        """
        Accessor function used to navigate through a labware by row.

        With indexing one can treat it as a typical python nested list.
        To access row A for example, simply write: labware.rows()[0]. This
        will output ['A1', 'A2', 'A3', 'A4'...]

        Note that this method takes args for backward-compatibility, but use
        of args is deprecated and will be removed in future versions. Args
        can be either strings or integers, but must all be the same type (e.g.:
        `self.rows(1, 4, 8)` or `self.rows('A', 'B')`, but  `self.rows('A', 4)`
        is invalid.

        :return: A list of row lists
        """
        grid = self._implementation.get_well_grid()
        if not args:
            res = grid.get_rows()
        elif isinstance(args[0], int):
            res = [grid.get_rows()[idx] for idx in args]
        elif isinstance(args[0], str):
            res = [grid.get_row(idx) for idx in args]
        else:
            raise TypeError
        return [[self._well_from_impl(w) for w in row] for row in res]

    @requires_version(2, 0)
    def rows_by_name(self) -> Dict[str, List[Well]]:
        """
        Accessor function used to navigate through a labware by row name.

        With indexing one can treat it as a typical python dictionary.
        To access row A for example, simply write: labware.rows_by_name()['A']
        This will output ['A1', 'A2', 'A3', 'A4'...].

        :return: Dictionary of Well lists keyed by row name
        """
        row_dict = self._implementation.get_well_grid().get_row_dict()
        return {k: [self._well_from_impl(w) for w in v] for k, v in row_dict.items()}

    @requires_version(2, 0)
    def rows_by_index(self) -> Dict[str, List[Well]]:
        """
        .. deprecated:: 2.0
            Use :py:meth:`rows_by_name` instead.
        """
        MODULE_LOG.warning("rows_by_index is deprecated. Use rows_by_name instead.")
        return self.rows_by_name()

    @requires_version(2, 0)
    def columns(self, *args) -> List[List[Well]]:
        """
        Accessor function used to navigate through a labware by column.

        With indexing one can treat it as a typical python nested list.
        To access row A for example,
        simply write: labware.columns()[0]
        This will output ['A1', 'B1', 'C1', 'D1'...].

        Note that this method takes args for backward-compatibility, but use
        of args is deprecated and will be removed in future versions. Args
        can be either strings or integers, but must all be the same type (e.g.:
        `self.columns(1, 4, 8)` or `self.columns('1', '2')`, but
        `self.columns('1', 4)` is invalid.

        :return: A list of column lists
        """
        grid = self._implementation.get_well_grid()
        if not args:
            res = grid.get_columns()
        elif isinstance(args[0], int):
            res = [grid.get_columns()[idx] for idx in args]
        elif isinstance(args[0], str):
            res = [grid.get_column(idx) for idx in args]
        else:
            raise TypeError
        return [[self._well_from_impl(w) for w in col] for col in res]

    @requires_version(2, 0)
    def columns_by_name(self) -> Dict[str, List[Well]]:
        """
        Accessor function used to navigate through a labware by column name.

        With indexing one can treat it as a typical python dictionary.
        To access row A for example,
        simply write: labware.columns_by_name()['1']
        This will output ['A1', 'B1', 'C1', 'D1'...].

        :return: Dictionary of Well lists keyed by column name
        """
        column_dict = self._implementation.get_well_grid().get_column_dict()
        return {k: [self._well_from_impl(w) for w in v] for k, v in column_dict.items()}

    @requires_version(2, 0)
    def columns_by_index(self) -> Dict[str, List[Well]]:
        """
        .. deprecated:: 2.0
            Use :py:meth:`columns_by_name` instead.
        """
        MODULE_LOG.warning(
            "columns_by_index is deprecated. Use columns_by_name instead."
        )
        return self.columns_by_name()

    @property  # type: ignore
    @requires_version(2, 0)
    def highest_z(self) -> float:
        """
        The z-coordinate of the tallest single point anywhere on the labware.

        This is drawn from the 'dimensions'/'zDimension' elements of the
        labware definition and takes into account the calibration offset.
        """
        return self._implementation.highest_z

    @property
    def _is_tiprack(self) -> bool:
        """as is_tiprack but not subject to version checking for speed"""
        return self._implementation.is_tiprack()

    @property  # type: ignore
    @requires_version(2, 0)
    def is_tiprack(self) -> bool:
        return self._is_tiprack

    @property  # type: ignore
    @requires_version(2, 0)
    def tip_length(self) -> float:
        return self._implementation.get_tip_length()

    @tip_length.setter
    def tip_length(self, length: float):
        self._implementation.set_tip_length(length)

    def next_tip(
        self, num_tips: int = 1, starting_tip: Optional[Well] = None
    ) -> Optional[Well]:
        """
        Find the next valid well for pick-up.

        Determines the next valid start tip from which to retrieve the
        specified number of tips. There must be at least `num_tips` sequential
        wells for which all wells have tips, in the same column.

        :param num_tips: target number of sequential tips in the same column
        :type num_tips: int
        :param starting_tip: The :py:class:`.Well` from which to start search.
                for an available tip.
        :type starting_tip: :py:class:`.Well`
        :return: the :py:class:`.Well` meeting the target criteria, or None
        """
        assert num_tips > 0, "Bad call to next_tip: num_tips <= 0"

        well = self._implementation.get_tip_tracker().next_tip(
            num_tips=num_tips, starting_tip=starting_tip._impl if starting_tip else None
        )
        return self._well_from_impl(well) if well else None

    def use_tips(self, start_well: Well, num_channels: int = 1):
        """
        Removes tips from the tip tracker.

        This method should be called when a tip is picked up. Generally, it
        will be called with `num_channels=1` or `num_channels=8` for single-
        and multi-channel respectively. If picking up with more than one
        channel, this method will automatically determine which tips are used
        based on the start well, the number of channels, and the geometry of
        the tiprack.

        :param start_well: The :py:class:`.Well` from which to pick up a tip.
                           For a single-channel pipette, this is the well to
                           send the pipette to. For a multi-channel pipette,
                           this is the well to send the back-most nozzle of the
                           pipette to.
        :type start_well: :py:class:`.Well`
        :param num_channels: The number of channels for the current pipette
        :type num_channels: int
        """
        assert num_channels > 0, "Bad call to use_tips: num_channels<=0"

        fail_if_full = self._api_version < APIVersion(2, 2)

        self._implementation.get_tip_tracker().use_tips(
            start_well=start_well._impl,
            num_channels=num_channels,
            fail_if_full=fail_if_full,
        )

    def __repr__(self):
        return self._implementation.get_display_name()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Labware):
            return NotImplemented
        return self._implementation == other._implementation

    def __hash__(self):
        return hash((self._implementation, self._api_version))

    def previous_tip(self, num_tips: int = 1) -> Optional[Well]:
        """
        Find the best well to drop a tip in.

        This is the well from which the last tip was picked up, if there's
        room. It can be used to return tips to the tip tracker.

        :param num_tips: target number of tips to return, sequential in a
                         column
        :type num_tips: int
        :return: The :py:class:`.Well` meeting the target criteria, or ``None``
        """
        # This logic is the inverse of :py:meth:`next_tip`
        assert num_tips > 0, "Bad call to previous_tip: num_tips <= 0"

        well = self._implementation.get_tip_tracker().previous_tip(num_tips=num_tips)
        return self._well_from_impl(well) if well else None

    def return_tips(self, start_well: Well, num_channels: int = 1):
        """
        Re-adds tips to the tip tracker

        This method should be called when a tip is dropped in a tiprack. It
        should be called with ``num_channels=1`` or ``num_channels=8`` for
        single- and multi-channel respectively. If returning more than one
        channel, this method will automatically determine which tips are
        returned based on the start well, the number of channels,
        and the tiprack geometry.

        Note that unlike :py:meth:`use_tips`, calling this method in a way
        that would drop tips into wells with tips in them will raise an
        exception; this should only be called on a valid return of
        :py:meth:`previous_tip`.

        :param start_well: The :py:class:`.Well` into which to return a tip.
        :type start_well: :py:class:`.Well`
        :param num_channels: The number of channels for the current pipette
        :type num_channels: int
        """
        # This logic is the inverse of :py:meth:`use_tips`
        assert num_channels > 0, "Bad call to return_tips: num_channels <= 0"

        self._implementation.get_tip_tracker().return_tips(
            start_well=start_well._impl, num_channels=num_channels
        )

    @requires_version(2, 0)
    def reset(self):
        """Reset all tips in a tiprack"""
        if self._is_tiprack:
            self._implementation.reset_tips()

    def _well_from_impl(self, well: WellImplementation) -> Well:
        return Well(well_implementation=well, api_level=self._api_version)


def save_definition(
    labware_def: "LabwareDefinition",
    force: bool = False,
    location: Optional[Path] = None,
) -> None:
    """
    Save a labware definition

    :param labware_def: A deserialized JSON labware definition
    :param bool force: If true, overwrite an existing definition if found.
        Cannot overwrite Opentrons definitions.
    :param location: The path of the labware definition.
    """
    labware_module.save_definition(
        labware_def=labware_def, force=force, location=location
    )


def verify_definition(
    contents: Union[AnyStr, "LabwareDefinition", Dict[str, Any]]
) -> "LabwareDefinition":
    """Verify that an input string is a labware definition and return it.

    If the definition is invalid, an exception is raised; otherwise parse the
    json and return the valid definition.

    :raises json.JsonDecodeError: If the definition is not valid json
    :raises jsonschema.ValidationError: If the definition is not valid.
    :returns: The parsed definition
    """
    return labware_module.verify_definition(contents=contents)


def get_labware_definition(
    load_name: str,
    namespace: Optional[str] = None,
    version: Optional[int] = None,
    bundled_defs: Dict[str, "LabwareDefinition"] = None,
    extra_defs: Dict[str, "LabwareDefinition"] = None,
) -> "LabwareDefinition":
    """
    Look up and return a definition by load_name + namespace + version and
        return it or raise an exception

    :param str load_name: corresponds to 'loadName' key in definition
    :param str namespace: The namespace the labware definition belongs to.
        If unspecified, will search 'opentrons' then 'custom_beta'
    :param int version: The version of the labware definition. If unspecified,
        will use version 1.
    :param bundled_defs: A bundle of labware definitions to exlusively use for
        finding labware definitions, if specified
    :param extra_defs: An extra set of definitions (in addition to the system
        definitions) in which to search
    """
    return labware_module.get_labware_definition(
        load_name=load_name,
        namespace=namespace,
        version=version,
        bundled_defs=bundled_defs,
        extra_defs=extra_defs,
    )


def get_all_labware_definitions() -> List[str]:
    """
    Return a list of standard and custom labware definitions with load_name +
        name_space + version existing on the robot
    """
    return labware_module.get_all_labware_definitions()


def split_tipracks(tip_racks: List[Labware]) -> Tuple[Labware, List[Labware]]:
    try:
        rest = tip_racks[1:]
    except IndexError:
        rest = []
    return tip_racks[0], rest


def select_tiprack_from_list(
    tip_racks: List[Labware], num_channels: int, starting_point: Optional[Well] = None
) -> Tuple[Labware, Well]:

    try:
        first, rest = split_tipracks(tip_racks)
    except IndexError:
        raise OutOfTipsError

    if starting_point and starting_point.parent != first:
        raise TipSelectionError(
            "The starting tip you selected " f"does not exist in {first}"
        )
    elif starting_point:
        first_well = starting_point
    else:
        first_well = first.wells()[0]

    next_tip = first.next_tip(num_channels, first_well)
    if next_tip:
        return first, next_tip
    else:
        return select_tiprack_from_list(rest, num_channels)


def select_tiprack_from_list_paired_pipettes(
    tip_racks: List[Labware],
    p_channels: int,
    s_channels: int,
    starting_point: Optional[Well] = None,
) -> Tuple[Labware, Well]:
    """
    Helper function utilized in :py:attr:`PairedInstrumentContext`
    to determine which pipette tiprack to pick up from.

    If a starting point is specified, this method with check
    that the parent of that tip was correctly filtered.

    If a starting point is not specified, this method will filter
    tipracks until it finds a well that is not empty.

    :return: A Tuple of the tiprack and well to move to. In this
    instance the starting well is specific to the primary pipette.
    :raises TipSelectionError: if the starting tip specified
    does not exist in the filtered tipracks.
    """
    try:
        first, rest = split_tipracks(tip_racks)
    except IndexError:
        raise OutOfTipsError

    if starting_point and starting_point.parent != first:
        raise TipSelectionError(
            "The starting tip you selected " f"does not exist in {first}"
        )
    elif starting_point:
        primary_well = starting_point
    else:
        primary_well = first.wells()[0]

    try:
        secondary_point = labware_column_shift(primary_well, first)
    except IndexError:
        return select_tiprack_from_list_paired_pipettes(rest, p_channels, s_channels)

    primary_next_tip = first.next_tip(p_channels, starting_point)
    secondary_next_tip = first.next_tip(s_channels, secondary_point)
    if primary_next_tip and secondary_next_tip:
        return first, primary_next_tip
    else:
        return select_tiprack_from_list_paired_pipettes(rest, p_channels, s_channels)


def filter_tipracks_to_start(
    starting_point: Well, tipracks: List[Labware]
) -> List[Labware]:
    return list(dropwhile(lambda tr: starting_point.parent != tr, tipracks))


def next_available_tip(
    starting_tip: Optional[Well], tip_racks: List[Labware], channels: int
) -> Tuple[Labware, Well]:
    start = starting_tip
    if start is None:
        return select_tiprack_from_list(tip_racks, channels)
    else:
        return select_tiprack_from_list(
            filter_tipracks_to_start(start, tip_racks), channels, start
        )


def get_labware_hash(labware: "Labware") -> str:
    return labware_module.get_labware_hash(labware._implementation)


def get_labware_hash_with_parent(labware: "Labware") -> str:
    return labware_module.get_labware_hash_with_parent(labware._implementation)


def load_from_definition(
    definition: "LabwareDefinition",
    parent: Location,
    label: Optional[str] = None,
    api_level: Optional[APIVersion] = None,
) -> Labware:
    """
    Return a labware object constructed from a provided labware definition dict

    :param definition: A dict representing all required data for a labware,
        including metadata such as the display name of the labware, a
        definition of the order to iterate over wells, the shape of wells
        (shape, physical dimensions, etc), and so on. The correct shape of
        this definition is governed by the "labware-designer" project in
        the Opentrons/opentrons repo.
    :param parent: A :py:class:`.Location` representing the location where
                   the front and left most point of the outside of labware is
                   (often the front-left corner of a slot on the deck).
    :param str label: An optional label that will override the labware's
                      display name from its definition
    :param APIVersion api_level: the API version to set for the loaded labware
                                 instance. The :py:class:`.Labware` will
                                 conform to this level. If not specified,
                                 defaults to :py:attr:`.MAX_SUPPORTED_VERSION`.
    """
    return Labware(
        implementation=labware_module.load_from_definition(
            definition=definition, parent=parent, label=label
        ),
        api_level=api_level,
    )


def save_calibration(labware: "Labware", delta: Point):
    labware_module.save_calibration(labware._implementation, delta)


def load(
    load_name: str,
    parent: Location,
    label: Optional[str] = None,
    namespace: Optional[str] = None,
    version: int = 1,
    bundled_defs: Optional[Dict[str, "LabwareDefinition"]] = None,
    extra_defs: Optional[Dict[str, "LabwareDefinition"]] = None,
    api_level: Optional[APIVersion] = None,
) -> Labware:
    """
    Return a labware object constructed from a labware definition dict looked
    up by name (definition must have been previously stored locally on the
    robot)

    :param load_name: A string to use for looking up a labware definition
        previously saved to disc. The definition file must have been saved in a
        known location
    :param parent: A :py:class:`.Location` representing the location where
                   the front and left most point of the outside of labware is
                   (often the front-left corner of a slot on the deck).
    :param str label: An optional label that will override the labware's
                      display name from its definition
    :param str namespace: The namespace the labware definition belongs to.
        If unspecified, will search 'opentrons' then 'custom_beta'
    :param int version: The version of the labware definition. If unspecified,
        will use version 1.
    :param bundled_defs: If specified, a mapping of labware names to labware
        definitions. Only the bundle will be searched for definitions.
    :param extra_defs: If specified, a mapping of labware names to labware
        definitions. If no bundle is passed, these definitions will also be
        searched.
    :param APIVersion api_level: the API version to set for the loaded labware
                                 instance. The :py:class:`.Labware` will
                                 conform to this level. If not specified,
                                 defaults to :py:attr:`.MAX_SUPPORTED_VERSION`.
    """

    return Labware(
        implementation=labware_module.load(
            load_name=load_name,
            parent=parent,
            label=label,
            namespace=namespace,
            version=version,
            bundled_defs=bundled_defs,
            extra_defs=extra_defs,
        ),
        api_level=api_level,
    )



"""ProtocolEngine class definition."""
from typing import Optional
from opentrons.hardware_control import API as HardwareAPI

from .resources import ModelUtils
from .commands import Command, CommandRequest, CommandMapper
from .execution import QueueWorker

from .state import (
    StateStore,
    StateView,
    PlayAction,
    PauseAction,
    StopAction,
    UpdateCommandAction,
)


class ProtocolEngine:
    """Main ProtocolEngine class.

    A ProtocolEngine instance holds the state of a protocol as it executes,
    and manages calls to a command executor that actually implements the logic
    of the commands themselves.
    """

    _hardware_api: HardwareAPI
    _state_store: StateStore
    _queue_worker: QueueWorker
    _command_mapper: CommandMapper
    _model_utils: ModelUtils

    def __init__(
        self,
        hardware_api: HardwareAPI,
        state_store: StateStore,
        queue_worker: QueueWorker,
        command_mapper: Optional[CommandMapper] = None,
        model_utils: Optional[ModelUtils] = None,
    ) -> None:
        """Initialize a ProtocolEngine instance.

        This constructor does not inject provider implementations. Prefer the
        ProtocolEngine.create factory classmethod.
        """
        self._hardware_api = hardware_api
        self._state_store = state_store
        self._queue_worker = queue_worker
        self._command_mapper = command_mapper or CommandMapper()
        self._model_utils = model_utils or ModelUtils()

    @property
    def state_view(self) -> StateView:
        """Get an interface to retrieve calculated state values."""
        return self._state_store

    def play(self) -> None:
        """Start or resume executing commands in the queue."""
        # TODO(mc, 2021-08-05): if starting, ensure plungers motors are
        # homed if necessary
        action = PlayAction()
        self._state_store.commands.validate_action_allowed(action)
        self._state_store.handle_action(action)
        self._queue_worker.start()

    def pause(self) -> None:
        """Pause executing commands in the queue."""
        action = PauseAction()
        self._state_store.commands.validate_action_allowed(action)
        self._state_store.handle_action(action)

    def add_command(self, request: CommandRequest) -> Command:
        """Add a command to the `ProtocolEngine`'s queue.

        Arguments:
            request: The command type and payload data used to construct
                the command in state.

        Returns:
            The full, newly queued command.
        """
        command = self._command_mapper.map_request_to_command(
            request=request,
            command_id=self._model_utils.generate_id(),
            created_at=self._model_utils.get_timestamp(),
        )
        self._state_store.handle_action(UpdateCommandAction(command=command))

        return command

    async def add_and_execute_command(self, request: CommandRequest) -> Command:
        """Add a command to the queue and wait for it to complete.

        The engine must be started by calling `play` before the command will
        execute. You only need to call `play` once.

        Arguments:
            request: The command type and payload data used to construct
                the command in state.

        Returns:
            The completed command, whether it succeeded or failed.
        """
        command = self.add_command(request)

        await self._state_store.wait_for(
            condition=self._state_store.commands.get_is_complete,
            command_id=command.id,
        )

        return self._state_store.commands.get(command_id=command.id)

    async def halt(self) -> None:
        """Halt execution, stopping all motion and cancelling future commands.

        You should call `stop` after calling `halt` for cleanup and to allow
        the engine to settle and recover.
        """
        self._state_store.handle_action(StopAction())
        self._queue_worker.cancel()
        await self._hardware_api.halt()

    async def stop(self, wait_until_complete: bool = False) -> None:
        """Gracefully stop the ProtocolEngine, waiting for it to become idle.

        The engine will finish executing its current command (if any),
        and then shut down. After an engine has been `stop`'ed, it cannot
        be restarted.

        This method should not raise, but if any exceptions happen during
        execution that are not properly caught by the CommandExecutor, they
        will be raised here.

        Arguments:
            wait_until_complete: Wait until _all_ commands have completed
                before stopping the engine.
        """
        if wait_until_complete:
            await self._state_store.wait_for(
                condition=self._state_store.commands.get_all_complete
            )

        self._state_store.handle_action(StopAction())

        try:
            await self._queue_worker.join()
        finally:
            await self._hardware_api.stop(home_after=False)



from pathlib import Path

from opentrons.config import get_opentrons_path

OPENTRONS_NAMESPACE = "opentrons"
CUSTOM_NAMESPACE = "custom_beta"
STANDARD_DEFS_PATH = Path("labware/definitions/2")
USER_DEFS_PATH = get_opentrons_path("labware_user_definitions_dir_v2")

SHORT_TRASH_DECK = "ot2_short_trash"
STANDARD_DECK = "ot2_standard"



"""Protocol engine state management."""
from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, List, Optional, Sequence, TypeVar

from opentrons_shared_data.deck.dev_types import DeckDefinitionV2

from ..resources import DeckFixedLabware
from .actions import Action
from .abstract_store import HasState, HandlesActions
from .change_notifier import ChangeNotifier
from .commands import CommandState, CommandStore, CommandView
from .labware import LabwareState, LabwareStore, LabwareView
from .pipettes import PipetteState, PipetteStore, PipetteView
from .geometry import GeometryView
from .motion import MotionView


ReturnT = TypeVar("ReturnT")


@dataclass(frozen=True)
class State:
    """Underlying engine state."""

    commands: CommandState
    labware: LabwareState
    pipettes: PipetteState


class StateView(HasState[State]):
    """A read-only view of computed state."""

    _state: State
    _commands: CommandView
    _labware: LabwareView
    _pipettes: PipetteView
    _geometry: GeometryView
    _motion: MotionView

    @property
    def commands(self) -> CommandView:
        """Get state view selectors for commands state."""
        return self._commands

    @property
    def labware(self) -> LabwareView:
        """Get state view selectors for labware state."""
        return self._labware

    @property
    def pipettes(self) -> PipetteView:
        """Get state view selectors for pipette state."""
        return self._pipettes

    @property
    def geometry(self) -> GeometryView:
        """Get state view selectors for derived geometry state."""
        return self._geometry

    @property
    def motion(self) -> MotionView:
        """Get state view selectors for derived motion state."""
        return self._motion


class StateStore(StateView, HandlesActions):
    """ProtocolEngine state store.

    A StateStore manages several substores, which will modify themselves in
    reaction to commands and other protocol events. State instances inside
    stores should be treated as immutable.
    """

    def __init__(
        self,
        deck_definition: DeckDefinitionV2,
        deck_fixed_labware: Sequence[DeckFixedLabware],
        change_notifier: Optional[ChangeNotifier] = None,
    ) -> None:
        """Initialize a StateStore and its substores.

        Arguments:
            deck_definition: The deck definition to preload into
                labware state.
            deck_fixed_labware: Labware definitions from the deck
                definition to preload into labware state.
            change_notifier: Internal state change notifier.
        """
        self._command_store = CommandStore()
        self._pipette_store = PipetteStore()
        self._labware_store = LabwareStore(
            deck_fixed_labware=deck_fixed_labware,
            deck_definition=deck_definition,
        )

        self._substores: List[HandlesActions] = [
            self._command_store,
            self._pipette_store,
            self._labware_store,
        ]

        self._change_notifier = change_notifier or ChangeNotifier()
        self._initialize_state()

    def handle_action(self, action: Action) -> None:
        """Modify State in reaction to an action.

        Arguments:
            action: An action object representing a state change. Will be
                passed to all substores so they can react accordingly.
        """
        for substore in self._substores:
            substore.handle_action(action)

        self._update_state_views()

    async def wait_for(
        self,
        condition: Callable[..., Optional[ReturnT]],
        *args: Any,
        **kwargs: Any,
    ) -> ReturnT:
        """Wait for a condition to become true, checking whenever state changes.

        !!! Warning:
            In general, callers should not trigger a state change via
            `handle_action` directly after a `wait_for`, as it may interfere
            with other subscribers. If you _must_ trigger a state change, ensure
            the change cannot affect the `condition`'s of other `wait_for`'s.

        Arguments:
            condition: A function that returns a truthy value when the `await`
                should resolve
            *args: Positional arguments to pass to `condition`
            **kwargs: Named arguments to pass to `condition`

        Returns:
            The truthy value returned by the `condition` function.
        """
        predicate = partial(condition, *args, **kwargs)
        is_done = predicate()

        while not is_done:
            await self._change_notifier.wait()
            is_done = predicate()

        return is_done

    def _get_next_state(self) -> State:
        """Get a new instance of the state value object."""
        return State(
            commands=self._command_store.state,
            labware=self._labware_store.state,
            pipettes=self._pipette_store.state,
        )

    def _initialize_state(self) -> None:
        """Initialize state data and view."""
        state = self._get_next_state()

        # Base states
        self._state = state
        self._commands = CommandView(state.commands)
        self._labware = LabwareView(state.labware)
        self._pipettes = PipetteView(state.pipettes)

        # Derived states
        self._geometry = GeometryView(labware_view=self._labware)
        self._motion = MotionView(
            labware_view=self._labware,
            pipette_view=self._pipettes,
            geometry_view=self._geometry,
        )

    def _update_state_views(self) -> None:
        """Update state view interfaces to use latest underlying values."""
        state = state = self._get_next_state()

        self._state = state
        self._commands._state = state.commands
        self._labware._state = state.labware
        self._pipettes._state = state.pipettes

        self._change_notifier.notify()



""" protocol_api: The user-facing API for OT2 protocols.

This package defines classes and functions for access through a protocol to
control the OT2.

"""
from opentrons.protocols.api_support.definitions import (
    MAX_SUPPORTED_VERSION,
    MIN_SUPPORTED_VERSION,
)
from . import labware
from .contexts import (
    ProtocolContext,
    InstrumentContext,
    PairedInstrumentContext,
    TemperatureModuleContext,
    MagneticModuleContext,
    ThermocyclerContext,
)

__all__ = [
    "MAX_SUPPORTED_VERSION",
    "MIN_SUPPORTED_VERSION",
    "ProtocolContext",
    "InstrumentContext",
    "TemperatureModuleContext",
    "MagneticModuleContext",
    "ThermocyclerContext",
    "PairedInstrumentContext",
    "labware",
]



from opentrons import config
import os
import logging

log = logging.getLogger(__name__)
lock_file_path = "/tmp/resin/resin-updates.lock"


def lock_updates():
    if config.IS_ROBOT:
        import fcntl

        try:
            with open(lock_file_path, "w") as fd:
                fd.write("a")
                fcntl.flock(fd, fcntl.LOCK_EX)
                fd.close()
        except OSError:
            log.warning("Unable to create resin-update lock file")


def unlock_updates():
    if config.IS_ROBOT and os.path.exists(lock_file_path):
        os.remove(lock_file_path)



"""Python protocol executor."""
import asyncio
from concurrent.futures import ThreadPoolExecutor
from functools import partial

from opentrons.protocol_api_experimental import ProtocolContext
from .python_file_reader import PythonProtocol


class PythonExecutor:
    """Execute a given PythonProtocol's run method with a ProtocolContext."""

    @staticmethod
    async def execute(protocol: PythonProtocol, context: ProtocolContext) -> None:
        """Execute a PythonProtocol using the given ProtocolContext.

        Runs the protocol asynchronously in a child thread.
        """
        loop = asyncio.get_running_loop()

        with ThreadPoolExecutor(max_workers=1) as executor:
            await loop.run_in_executor(
                executor=executor,
                func=partial(protocol.run, context),
            )



import asyncio
import logging
from typing import Optional, List, Dict, Mapping
from dataclasses import dataclass
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.types import ThermocyclerLidStatus, Temperature, PlateTemperature
from opentrons.hardware_control.modules.lid_temp_status import LidTemperatureStatus
from opentrons.hardware_control.modules.plate_temp_status import PlateTemperatureStatus
from opentrons.hardware_control.modules.types import TemperatureStatus
from opentrons.hardware_control.poller import Reader, WaitableListener, Poller

from ..execution_manager import ExecutionManager
from . import types, update, mod_abc
from opentrons.drivers.thermocycler import (
    AbstractThermocyclerDriver,
    SimulatingDriver,
    ThermocyclerDriver,
)


MODULE_LOG = logging.getLogger(__name__)

POLLING_FREQUENCY_SEC = 1.0
SIM_POLLING_FREQUENCY_SEC = POLLING_FREQUENCY_SEC / 20.0

TEMP_UPDATE_RETRIES = 50


class ThermocyclerError(Exception):
    pass


class Thermocycler(mod_abc.AbstractModule):
    """
    Under development. API subject to change without a version bump
    """

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        interrupt_callback: types.InterruptCallback = None,
        simulating: bool = False,
        loop: asyncio.AbstractEventLoop = None,
        sim_model: str = None,
        **kwargs,
    ):
        """
        Build and connect to a Thermocycler

        Args:
            port: The port to connect to
            usb_port: USB Port
            execution_manager: Execution manager.
            interrupt_callback: Optional interrupt callback
            simulating: whether to build a simulating driver
            loop: Loop
            sim_model: The model name used by simulator
            **kwargs: Module specific values.
                can be 'polling_frequency' to specify the polling frequency in
                seconds

        Returns:
            Thermocycler instance.
        """
        polling_frequency = kwargs.get("polling_frequency")
        driver: AbstractThermocyclerDriver
        if not simulating:
            driver = await ThermocyclerDriver.create(port=port, loop=loop)
            polling_frequency = polling_frequency or POLLING_FREQUENCY_SEC
        else:
            driver = SimulatingDriver()
            polling_frequency = polling_frequency or SIM_POLLING_FREQUENCY_SEC

        mod = cls(
            port=port,
            usb_port=usb_port,
            driver=driver,
            device_info=await driver.get_device_info(),
            interrupt_callback=interrupt_callback,
            loop=loop,
            execution_manager=execution_manager,
            polling_interval_sec=polling_frequency,
        )
        return mod

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        driver: AbstractThermocyclerDriver,
        device_info: Dict[str, str],
        interrupt_callback: types.InterruptCallback = None,
        loop: asyncio.AbstractEventLoop = None,
        polling_interval_sec: float = POLLING_FREQUENCY_SEC,
    ) -> None:
        """
        Constructor

        Args:
            port: The port the thermocycler is connected to.
            usb_port: The USB port.
            execution_manager: The hardware execution manager.
            driver: The thermocycler driver.
            device_info: The thermocycler device info.
            interrupt_callback: Optional interrupt callback.
            loop: Optional loop.
            polling_interval_sec: How often to poll thermocycler for status
        """
        super().__init__(
            port=port, usb_port=usb_port, loop=loop, execution_manager=execution_manager
        )
        self._driver = driver
        self._device_info = device_info
        self._listener = ThermocyclerListener(
            loop=loop, interrupt_callback=interrupt_callback
        )
        self._poller = Poller(
            interval_seconds=polling_interval_sec,
            listener=self._listener,
            reader=PollerReader(driver=self._driver),
            loop=loop,
        )
        self._hold_time_fuzzy_seconds = polling_interval_sec * 5
        self._interrupt_cb = interrupt_callback

        self._total_cycle_count: Optional[int] = None
        self._current_cycle_index: Optional[int] = None
        self._total_step_count: Optional[int] = None
        self._current_step_index: Optional[int] = None

    async def cleanup(self) -> None:
        """Stop the poller task."""
        await self._poller.stop_and_wait()

    @classmethod
    def name(cls) -> str:
        return "thermocycler"

    def model(self) -> str:
        return "thermocyclerModuleV1"

    @classmethod
    def bootloader(cls) -> types.UploadFunction:
        return update.upload_via_bossa

    def _clear_cycle_counters(self) -> None:
        """Clear the cycle counters."""
        self._total_cycle_count = None
        self._current_cycle_index = None
        self._total_step_count = None
        self._current_step_index = None

    async def deactivate_lid(self) -> None:
        """Deactivate the lid heating pad"""
        await self.wait_for_is_running()
        return await self._driver.deactivate_lid()

    async def deactivate_block(self) -> None:
        """Deactivate the block peltiers"""
        await self.wait_for_is_running()
        self._clear_cycle_counters()
        return await self._driver.deactivate_block()

    async def deactivate(self) -> None:
        """Deactivate the block peltiers and lid heating pad"""
        await self.wait_for_is_running()
        self._clear_cycle_counters()
        return await self._driver.deactivate_all()

    async def open(self) -> str:
        """Open the lid if it is closed"""
        await self.wait_for_is_running()
        await self._driver.open_lid()
        await self._wait_for_lid_status(ThermocyclerLidStatus.OPEN)
        return ThermocyclerLidStatus.OPEN

    async def close(self) -> str:
        """Close the lid if it is open"""
        await self.wait_for_is_running()
        await self._driver.close_lid()
        await self._wait_for_lid_status(ThermocyclerLidStatus.CLOSED)
        return ThermocyclerLidStatus.CLOSED

    def hold_time_probably_set(self, new_hold_time: Optional[float]) -> bool:
        """
        Since we can only get hold time *remaining* from TC, by the time we
        read hold_time after a set_temperature, the hold_time in TC could have
        started counting down. So instead of checking for equality, we will
        have to check if the hold_time returned from TC is within a few seconds
        of the new hold time. The number of seconds is determined by status
        polling frequency.
        """
        if new_hold_time is None:
            return True
        hold_time = self.hold_time
        if hold_time is None:
            return False
        lower_bound = max(0.0, new_hold_time - self._hold_time_fuzzy_seconds)
        return lower_bound <= hold_time <= new_hold_time

    async def set_temperature(
        self,
        temperature: float,
        hold_time_seconds: Optional[float] = None,
        hold_time_minutes: Optional[float] = None,
        ramp_rate: Optional[float] = None,
        volume: Optional[float] = None,
    ) -> None:
        """
        Set the temperature and wait.

        If hold time is set this function will return after
        the hold time expires.

        Otherwise it will return when the target temperature is reached.

        Args:
            temperature: The target temperature.
            hold_time_seconds: Optional number of seconds to wait.
            hold_time_minutes: Optional number of minutes to wait.
            ramp_rate: Optional ramp rate.
            volume: Optional volume.

        Returns: None
        """
        await self.wait_for_is_running()
        seconds = hold_time_seconds if hold_time_seconds is not None else 0
        minutes = hold_time_minutes if hold_time_minutes is not None else 0
        total_seconds = seconds + (minutes * 60)
        hold_time = total_seconds if total_seconds > 0 else 0
        if ramp_rate is not None:
            await self._driver.set_ramp_rate(ramp_rate=ramp_rate)
        await self._driver.set_plate_temperature(
            temp=temperature, hold_time=hold_time, volume=volume
        )

        # Wait for target temperature to be set.
        retries = 0
        while self.target != temperature or not self.hold_time_probably_set(hold_time):
            # Wait for the poller to update
            await self.wait_next_poll()
            retries += 1
            if retries > TEMP_UPDATE_RETRIES:
                raise ThermocyclerError(
                    f"Thermocycler driver set the block temp to "
                    f"T={temperature} & H={hold_time} but status reads "
                    f"T={self.target} & H={self.hold_time}"
                )

        if hold_time:
            task = self._loop.create_task(self._wait_for_hold(hold_time))
        else:
            task = self._loop.create_task(self._wait_for_temp())
        await self.make_cancellable(task)
        await task

    async def cycle_temperatures(
        self,
        steps: List[types.ThermocyclerStep],
        repetitions: int,
        volume: float = None,
    ) -> None:
        """
        Begin a set temperature cycle.

        Args:
            steps: The set temperature steps.
            repetitions: Number of repetitions.
            volume: Optional volume.

        Returns: None
        """
        await self.wait_for_is_running()
        self._total_cycle_count = repetitions
        self._total_step_count = len(steps)

        task = self._loop.create_task(self._execute_cycles(steps, repetitions, volume))
        await self.make_cancellable(task)
        await task

    async def set_lid_temperature(self, temperature: float) -> None:
        """Set the lid temperature in deg Celsius"""
        await self.wait_for_is_running()
        await self._driver.set_lid_temperature(temp=temperature)
        # Wait for target to be set
        retries = 0
        while self.lid_target != temperature:
            # Wait for the poller to update
            await self.wait_next_poll()
            retries += 1
            if retries > TEMP_UPDATE_RETRIES:
                raise ThermocyclerError(
                    f"Thermocycler driver set the lid temp to T={temperature}"
                    f"but status reads T={self.lid_target}"
                )
        task = self._loop.create_task(self._wait_for_lid_temp())
        await self.make_cancellable(task)
        await task

    async def _wait_for_lid_temp(self) -> None:
        """
        This method only exits if lid target temperature has been reached.

        Subject to change without a version bump.
        """
        if self.is_simulated:
            return

        while self._listener.lid_status != TemperatureStatus.HOLDING:
            await self._listener.wait_next_poll()

    async def _wait_for_temp(self) -> None:
        """
        This method only exits if set temperature has been reached.

        Subject to change without a version bump.
        """
        while self._listener.plate_status != TemperatureStatus.HOLDING:
            await self.wait_next_poll()

    async def _wait_for_hold(self, hold_time: float = 0) -> None:
        """
        This method returns only when hold time has elapsed
        """
        if self.is_simulated:
            return

        # If hold time is within the _hold_time_fuzzy_seconds time gap, then,
        # because of the driver's status poller delays, it is impossible to
        # know for certain if self.hold_time holds the most recent value.
        # So instead of counting on the cached self.hold_time, it is better to
        # just wait for hold_time time. (Skip if hold_time = 0 since we don't
        # want to wait in that case. Cached self.hold_time would be 0 anyway)
        if 0 < hold_time <= self._hold_time_fuzzy_seconds:
            await asyncio.sleep(hold_time)
        else:
            while self.hold_time != 0:
                await self.wait_next_poll()

    async def _wait_for_lid_status(self, status: ThermocyclerLidStatus) -> None:
        """Wait for lid status to be status."""
        while self.lid_status != status:
            await self.wait_next_poll()

    async def wait_next_poll(self) -> None:
        """Wait for the next poll to complete."""
        await self._listener.wait_next_poll()

    @property
    def lid_target(self) -> Optional[float]:
        return (
            self._listener.state.lid_temperature.target
            if self._listener.state
            else None
        )

    @property
    def lid_temp(self) -> Optional[float]:
        return (
            self._listener.state.lid_temperature.current
            if self._listener.state
            else None
        )

    @property
    def lid_status(self) -> Optional[str]:
        return self._listener.state.lid_status if self._listener.state else None

    @property
    def lid_temp_status(self) -> Optional[str]:
        return self._listener.lid_status

    @property
    def ramp_rate(self) -> Optional[float]:
        """Not supported."""
        return None

    @property
    def hold_time(self) -> Optional[float]:
        return (
            self._listener.state.plate_temperature.hold
            if self._listener.state
            else None
        )

    @property
    def temperature(self) -> Optional[float]:
        return (
            self._listener.state.plate_temperature.current
            if self._listener.state
            else None
        )

    @property
    def target(self) -> Optional[float]:
        return (
            self._listener.state.plate_temperature.target
            if self._listener.state
            else None
        )

    @property
    def status(self) -> str:
        return self._listener.plate_status

    @property
    def device_info(self) -> Mapping[str, str]:
        return self._device_info

    @property
    def total_cycle_count(self) -> Optional[int]:
        return self._total_cycle_count

    @property
    def current_cycle_index(self) -> Optional[int]:
        return self._current_cycle_index

    @property
    def total_step_count(self) -> Optional[int]:
        return self._total_step_count

    @property
    def current_step_index(self) -> Optional[int]:
        return self._current_step_index

    @property
    def live_data(self):
        return {
            "status": self.status,
            "data": {
                "lid": self.lid_status,
                "lidTarget": self.lid_target,
                "lidTemp": self.lid_temp,
                "currentTemp": self.temperature,
                "targetTemp": self.target,
                "holdTime": self.hold_time,
                "rampRate": self.ramp_rate,
                "currentCycleIndex": self.current_cycle_index,
                "totalCycleCount": self.total_cycle_count,
                "currentStepIndex": self.current_step_index,
                "totalStepCount": self.total_step_count,
            },
        }

    @property
    def is_simulated(self):
        return isinstance(self._driver, SimulatingDriver)

    async def prep_for_update(self):
        await self._driver.enter_programming_mode()

        new_port = await update.find_bootloader_port()

        return new_port or self.port

    async def _execute_cycle_step(
        self, step: types.ThermocyclerStep, volume: Optional[float]
    ) -> None:
        """
        Execute a thermocycler step.

        Args:
            step: The set temperature parameters.
            volume: The volume

        Returns: None
        """
        await self.wait_for_is_running()

        temperature = step.get("temperature")
        hold_time_minutes = step.get("hold_time_minutes", None)
        hold_time_seconds = step.get("hold_time_seconds", None)
        ramp_rate = step.get("ramp_rate", None)
        await self.set_temperature(
            temperature=temperature,  # type: ignore
            hold_time_minutes=hold_time_minutes,
            hold_time_seconds=hold_time_seconds,
            ramp_rate=ramp_rate,
            volume=volume,
        )

    async def _execute_cycles(
        self,
        steps: List[types.ThermocyclerStep],
        repetitions: int,
        volume: Optional[float] = None,
    ) -> None:
        """
        Execute cycles.

        Args:
            steps: The set temperature steps.
            repetitions: The number of repetitions
            volume: The optional volume.

        Returns: None
        """
        for rep in range(repetitions):
            self._current_cycle_index = rep + 1  # science starts at 1
            for step_idx, step in enumerate(steps):
                self._current_step_index = step_idx + 1  # science starts at 1
                await self._execute_cycle_step(step, volume)
                await self._wait_for_hold()


@dataclass
class PolledData:
    lid_status: ThermocyclerLidStatus
    lid_temperature: Temperature
    plate_temperature: PlateTemperature


class PollerReader(Reader[PolledData]):
    """Polled data reader."""

    def __init__(self, driver: AbstractThermocyclerDriver) -> None:
        """Constructor."""
        self._driver = driver

    async def read(self) -> PolledData:
        """Poll the thermocycler."""
        lid_status = await self._driver.get_lid_status()
        lid_temperature = await self._driver.get_lid_temperature()
        plate_temperature = await self._driver.get_plate_temperature()
        return PolledData(
            lid_status=lid_status,
            lid_temperature=lid_temperature,
            plate_temperature=plate_temperature,
        )


class ThermocyclerListener(WaitableListener[PolledData]):
    """Thermocycler state listener."""

    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        interrupt_callback: types.InterruptCallback = None,
    ) -> None:
        """Constructor."""
        super().__init__(loop=loop)
        self._callback = interrupt_callback
        self._polled_data: Optional[PolledData] = None
        self._plate_temperature_status = PlateTemperatureStatus()
        self._lid_temperature_status = LidTemperatureStatus()

    @property
    def state(self) -> Optional[PolledData]:
        return self._polled_data

    @property
    def plate_status(self) -> TemperatureStatus:
        return self._plate_temperature_status.status

    @property
    def lid_status(self) -> TemperatureStatus:
        return self._lid_temperature_status.status

    def on_poll(self, result: PolledData) -> None:
        """On new poll."""
        self._polled_data = result
        self._plate_temperature_status.update(result.plate_temperature)
        self._lid_temperature_status.update(result.lid_temperature)
        return super().on_poll(result)

    def on_error(self, exc: Exception) -> None:
        """On error."""
        if self._callback:
            self._callback(str(exc))



from __future__ import annotations

import asyncio
import logging
from typing import Optional

from opentrons.drivers.command_builder import CommandBuilder

from .errors import NoResponse, AlarmResponse, ErrorResponse
from .async_serial import AsyncSerial

log = logging.getLogger(__name__)


class SerialConnection:
    @classmethod
    async def create(
        cls,
        port: str,
        baud_rate: int,
        timeout: float,
        ack: str,
        name: Optional[str] = None,
        retry_wait_time_seconds: float = 0.1,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> SerialConnection:
        """
        Create a connection.

        Args:
            port: url or port to connect to
            baud_rate: baud rate
            timeout: timeout in seconds
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
            loop: optional event loop.

        Returns: SerialConnection
        """
        serial = await AsyncSerial.create(
            port=port, baud_rate=baud_rate, timeout=timeout, loop=loop
        )
        name = name or port
        return cls(
            serial=serial,
            port=port,
            name=name,
            ack=ack,
            retry_wait_time_seconds=retry_wait_time_seconds,
        )

    def __init__(
        self,
        serial: AsyncSerial,
        port: str,
        name: str,
        ack: str,
        retry_wait_time_seconds: float,
    ) -> None:
        """
        Constructor

        Args:
            serial: AsyncSerial object
            port: url or port to connect to
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
        """
        self._serial = serial
        self._port = port
        self._name = name
        self._ack = ack.encode()
        self._retry_wait_time_seconds = retry_wait_time_seconds
        self._send_data_lock = asyncio.Lock()

    async def send_command(
        self, command: CommandBuilder, retries: int = 0, timeout: Optional[float] = None
    ) -> str:
        """
        Send a command and return the response.

        Args:
            command: A command builder.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        return await self.send_data(
            data=command.build(), retries=retries, timeout=timeout
        )

    async def send_data(
        self, data: str, retries: int = 0, timeout: Optional[float] = None
    ) -> str:
        """
        Send data and return the response.

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        async with self._send_data_lock:
            return await self._send_data(data=data, retries=retries, timeout=timeout)

    async def _send_data(
        self, data: str, retries: int = 0, timeout: Optional[float] = None
    ) -> str:
        """
        Send data and return the response.

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        data_encode = data.encode()

        for retry in range(retries + 1):
            log.debug(f"{self.name}: Write -> {data_encode!r}")
            await self._serial.write(data=data_encode)

            response = await self._serial.read_until(match=self._ack, timeout=timeout)
            log.debug(f"{self.name}: Read <- {response!r}")

            if self._ack in response:
                # Remove ack from response
                response = response.replace(self._ack, b"")
                str_response = self.process_raw_response(
                    command=data, response=response.decode()
                )
                self.raise_on_error(response=str_response)
                return str_response

            log.info(f"{self.name}: retry number {retry}/{retries}")

            await self.on_retry()

        raise NoResponse(port=self._port, command=data)

    async def open(self) -> None:
        """Open the connection."""
        await self._serial.open()

    async def close(self) -> None:
        """Close the connection."""
        await self._serial.close()

    async def is_open(self) -> bool:
        """Check if connection is open."""
        return await self._serial.is_open()

    @property
    def port(self) -> str:
        return self._port

    @property
    def name(self) -> str:
        return self._name

    def raise_on_error(self, response: str) -> None:
        """
        Raise an error if the response contains an error

        Args:
            response: response

        Returns: None

        Raises: SerialException
        """
        lower = response.lower()
        if "error" in lower:
            raise ErrorResponse(port=self._port, response=response)

        if "alarm" in lower:
            raise AlarmResponse(port=self._port, response=response)

    async def on_retry(self) -> None:
        """
        Opportunity for derived classes to perform action between retries. Default
        behaviour is to wait then re-open the connection.

        Returns: None
        """
        await asyncio.sleep(self._retry_wait_time_seconds)
        await self._serial.close()
        await self._serial.open()

    def process_raw_response(self, command: str, response: str) -> str:
        """
        Opportunity for derived classes to process the raw response. Default
         strips white space.

        Args:
            command: The sent command.
            response: The raw read response minus ack.

        Returns:
            processed response.
        """
        return response.strip()



from typing import Dict

from .abstract import AbstractMagDeckDriver


MAG_DECK_MODELS = {
    "magneticModuleV1": "mag_deck_v1.1",
    "magneticModuleV2": "mag_deck_v20",
}


class SimulatingDriver(AbstractMagDeckDriver):
    def __init__(self, sim_model: str = None):
        self._height = 0.0
        self._model = MAG_DECK_MODELS[sim_model] if sim_model else "mag_deck_v1.1"

    async def probe_plate(self):
        pass

    async def home(self):
        pass

    async def move(self, location: float):
        self._height = location

    async def get_device_info(self) -> Dict[str, str]:
        return {
            "serial": "dummySerialMD",
            "model": self._model,
            "version": "dummyVersionMD",
        }

    async def connect(self):
        pass

    async def disconnect(self):
        pass

    async def enter_programming_mode(self):
        pass

    async def is_connected(self) -> bool:
        return True

    async def get_plate_height(self) -> float:
        return self._height

    async def get_mag_position(self) -> float:
        return self._height



import asyncio
import os
from pathlib import Path
from opentrons.config import IS_OSX


class CameraException(Exception):
    pass


async def take_picture(filename: Path, loop: asyncio.AbstractEventLoop = None):
    """Take a picture and save it to filename

    :param filename: Name of file to save picture to
    :param loop: optional loop to use
    :return: None
    :raises: CameraException
    """
    try:
        os.remove(filename)
    except OSError:
        pass

    cmd = "ffmpeg -f video4linux2 -s 640x480 -i /dev/video0 -ss 0:0:1 -frames 1"

    if IS_OSX:
        # Purely for development on macos
        cmd = 'ffmpeg -f avfoundation -framerate 1  -s 640x480  -i "0" -ss 0:0:1 -frames 1'  # noqa: E501

    proc = await asyncio.create_subprocess_shell(
        f"{cmd} {filename}",
        stderr=asyncio.subprocess.PIPE,
        loop=loop or asyncio.get_event_loop(),
    )

    res = await proc.stderr.read()  # type: ignore
    res = res.decode().strip()
    await proc.wait()

    if proc.returncode != 0:
        raise CameraException(res)
    if not filename.exists():
        raise CameraException("picture not saved")



from dataclasses import dataclass
from typing import Dict, Tuple
from typing_extensions import TypedDict


class AxisDict(TypedDict):
    X: float
    Y: float
    Z: float
    A: float
    B: float
    C: float


class CurrentDictDefault(TypedDict):
    default: AxisDict


CurrentDictModelEntries = TypedDict(
    "CurrentDictModelEntries",
    {"2.1": AxisDict, "A": AxisDict, "B": AxisDict, "C": AxisDict},
    total=False,
)


class CurrentDict(CurrentDictDefault, CurrentDictModelEntries):
    pass


@dataclass
class RobotConfig:
    name: str
    version: int
    gantry_steps_per_mm: Dict[str, float]
    acceleration: Dict[str, float]
    serial_speed: int
    default_pipette_configs: Dict[str, float]
    default_current: CurrentDict
    low_current: CurrentDict
    high_current: CurrentDict
    default_max_speed: AxisDict
    log_level: str
    z_retract_distance: float
    left_mount_offset: Tuple[float, float, float]



"""Load pipette command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import Optional, Type
from typing_extensions import Literal

from opentrons.types import MountType

from ..types import PipetteName
from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest

LoadPipetteCommandType = Literal["loadPipette"]


class LoadPipetteData(BaseModel):
    """Data needed to load a pipette on to a mount."""

    pipetteName: PipetteName = Field(
        ...,
        description="The load name of the pipette to be required.",
    )
    mount: MountType = Field(
        ...,
        description="The mount the pipette should be present on.",
    )
    pipetteId: Optional[str] = Field(
        None,
        description="An optional ID to assign to this pipette. If None, an ID "
        "will be generated.",
    )


class LoadPipetteResult(BaseModel):
    """Result data for executing a LoadPipette."""

    pipetteId: str = Field(
        ...,
        description="An ID to reference this pipette in subsequent commands.",
    )


class LoadPipetteImplementation(
    AbstractCommandImpl[LoadPipetteData, LoadPipetteResult]
):
    """Load pipette command implementation."""

    async def execute(self, data: LoadPipetteData) -> LoadPipetteResult:
        """Check that requested pipette is attached and assign its identifier."""
        loaded_pipette = await self._equipment.load_pipette(
            pipette_name=data.pipetteName,
            mount=data.mount,
            pipette_id=data.pipetteId,
        )

        return LoadPipetteResult(pipetteId=loaded_pipette.pipette_id)


class LoadPipette(BaseCommand[LoadPipetteData, LoadPipetteResult]):
    """Load pipette command model."""

    commandType: LoadPipetteCommandType = "loadPipette"
    data: LoadPipetteData
    result: Optional[LoadPipetteResult]

    _ImplementationCls: Type[LoadPipetteImplementation] = LoadPipetteImplementation


class LoadPipetteRequest(BaseCommandRequest[LoadPipetteData]):
    """Load pipette command creation request model."""

    commandType: LoadPipetteCommandType = "loadPipette"
    data: LoadPipetteData

    _CommandCls: Type[LoadPipette] = LoadPipette



from typing import Optional

from opentrons import types
from opentrons.hardware_control.types import PipettePair
from opentrons.protocols.api_support.types import APIVersion
from opentrons.hardware_control import CriticalPoint
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.api_support.util import (
    Clearances,
    build_edges,
    FlowRates,
    PlungerSpeeds,
)
from opentrons.protocols.context.paired_instrument import AbstractPairedInstrument
from opentrons.protocols.context.protocol_api.paired_instrument import PairedInstrument
from opentrons.protocols.geometry import planning
from opentrons.protocols.context.instrument import AbstractInstrument
from opentrons.protocols.context.protocol import AbstractProtocol
from opentrons.protocols.context.well import WellImplementation


class InstrumentContextImplementation(AbstractInstrument):
    """Implementation of the InstrumentContext interface."""

    _api_version: APIVersion
    _protocol_interface: AbstractProtocol
    _mount: types.Mount
    _instrument_name: str
    _default_speed: float
    _well_bottom_clearances: Clearances
    _flow_rates: FlowRates
    _speeds: PlungerSpeeds

    def __init__(
        self,
        protocol_interface: AbstractProtocol,
        mount: types.Mount,
        instrument_name: str,
        default_speed: float,
        api_version: Optional[APIVersion] = None,
    ):
        """ "Constructor"""
        self._api_version = api_version or MAX_SUPPORTED_VERSION
        self._protocol_interface = protocol_interface
        self._mount = mount
        self._instrument_name = instrument_name
        self._default_speed = default_speed
        self._well_bottom_clearances = Clearances(
            default_aspirate=1.0, default_dispense=1.0
        )
        self._flow_rates = FlowRates(self)
        self._speeds = PlungerSpeeds(self)
        self._flow_rates.set_defaults(api_level=self._api_version)

    def get_default_speed(self) -> float:
        """Gets the speed at which the robot's gantry moves."""
        return self._default_speed

    def set_default_speed(self, speed: float) -> None:
        """Sets the speed at which the robot's gantry moves."""
        self._default_speed = speed

    def aspirate(self, volume: float, rate: float) -> None:
        """Aspirate a given volume of liquid from the specified location, using
        this pipette."""
        self._protocol_interface.get_hardware().hardware.aspirate(
            self._mount, volume, rate
        )

    def dispense(self, volume: float, rate: float) -> None:
        """Dispense a volume of liquid (in microliters/uL) using this pipette
        into the specified location."""
        self._protocol_interface.get_hardware().hardware.dispense(
            self._mount, volume, rate
        )

    def blow_out(self) -> None:
        """Blow liquid out of the tip."""
        self._protocol_interface.get_hardware().hardware.blow_out(self._mount)

    def touch_tip(
        self, location: WellImplementation, radius: float, v_offset: float, speed: float
    ) -> None:
        """
        Touch the pipette tip to the sides of a well, with the intent of
        removing left-over droplets
        """
        # TODO al 20201110 - build_edges relies on where being a Well. This is
        #  an unpleasant compromise until refactoring build_edges to support
        #  WellImplementation.
        #  Also, build_edges should not require api_version.
        from opentrons.protocol_api.labware import Well

        edges = build_edges(
            where=Well(well_implementation=location),
            offset=v_offset,
            mount=self._mount,
            deck=self._protocol_interface.get_deck(),
            radius=radius,
            version=self._api_version,
        )
        for edge in edges:
            self._protocol_interface.get_hardware().hardware.move_to(
                self._mount, edge, speed
            )

    def pick_up_tip(
        self,
        well: WellImplementation,
        tip_length: float,
        presses: Optional[int],
        increment: Optional[float],
    ) -> None:
        """Pick up a tip for the pipette to run liquid-handling commands."""
        hw = self._protocol_interface.get_hardware().hardware
        geometry = well.get_geometry()

        hw.set_current_tiprack_diameter(self._mount, geometry.diameter)

        hw.pick_up_tip(self._mount, tip_length, presses, increment)
        hw.set_working_volume(self._mount, geometry.max_volume)

    def drop_tip(self, home_after: bool) -> None:
        """Drop the tip."""
        self._protocol_interface.get_hardware().hardware.drop_tip(
            self._mount, home_after=home_after
        )

    def home(self) -> None:
        """Home the mount"""
        self._protocol_interface.get_hardware().hardware.home_z(self._mount)
        self.home_plunger()

    def home_plunger(self) -> None:
        """Home the plunger associated with this mount."""
        self._protocol_interface.get_hardware().hardware.home_plunger(self._mount)

    def delay(self) -> None:
        """Delay protocol execution."""
        self._protocol_interface.delay(seconds=0, msg=None)

    def move_to(
        self,
        location: types.Location,
        force_direct: bool,
        minimum_z_height: Optional[float],
        speed: Optional[float],
    ) -> None:
        """Move the instrument."""
        # prevent direct movement bugs in PAPI version >= 2.10
        location_cache_mount = (
            self._mount if self._api_version >= APIVersion(2, 10) else None
        )

        last_location = self._protocol_interface.get_last_location(
            mount=location_cache_mount
        )

        if last_location:
            from_lw = last_location.labware
        else:
            from_lw = LabwareLike(None)

        if not speed:
            speed = self.get_default_speed()

        hardware = self._protocol_interface.get_hardware().hardware

        from_center = from_lw.center_multichannel_on_wells() if from_lw else False
        cp_override = CriticalPoint.XY_CENTER if from_center else None

        from_loc = types.Location(
            hardware.gantry_position(self._mount, critical_point=cp_override), from_lw
        )

        instr_max_height = hardware.get_instrument_max_height(self._mount)
        moves = planning.plan_moves(
            from_loc,
            location,
            self._protocol_interface.get_deck(),
            instr_max_height,
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
        )

        try:
            for move in moves:
                hardware.move_to(
                    self._mount,
                    move[0],
                    critical_point=move[1],
                    speed=speed,
                    max_speeds=self._protocol_interface.get_max_speeds().data,
                )
        except Exception:
            self._protocol_interface.set_last_location(None)
            raise
        else:
            self._protocol_interface.set_last_location(
                location=location, mount=location_cache_mount
            )

    def get_mount(self) -> types.Mount:
        """Get the mount this pipette is attached to."""
        return self._mount

    def get_instrument_name(self) -> str:
        """Get the instrument name."""
        return self._instrument_name

    def get_pipette_name(self) -> str:
        """Get the pipette name."""
        return self.get_pipette()["name"]

    def get_model(self) -> str:
        """Get the model name."""
        return self.get_pipette()["model"]

    def get_min_volume(self) -> float:
        """Get the min volume."""
        return self.get_pipette()["min_volume"]

    def get_max_volume(self) -> float:
        """Get the max volume."""
        return self.get_pipette()["max_volume"]

    def get_current_volume(self) -> float:
        """Get the current volume."""
        return self.get_pipette()["current_volume"]

    def get_available_volume(self) -> float:
        """Get the available volume."""
        return self.get_pipette()["available_volume"]

    def get_pipette(self) -> PipetteDict:
        """Get the hardware pipette dictionary."""
        hw_manager = self._protocol_interface.get_hardware()
        pipette = hw_manager.hardware.get_attached_instrument(self._mount)
        if pipette is None:
            raise types.PipetteNotAttachedError()
        return pipette

    def get_channels(self) -> int:
        """Number of channels."""
        return self.get_pipette()["channels"]

    def has_tip(self) -> bool:
        """Whether a tip is attached."""
        return self.get_pipette()["has_tip"]

    def is_ready_to_aspirate(self) -> bool:
        return self.get_pipette()["ready_to_aspirate"]

    def prepare_for_aspirate(self) -> None:
        self._protocol_interface.get_hardware().hardware.prepare_for_aspirate(
            self._mount
        )

    def get_return_height(self) -> float:
        """The height to return a tip to its tiprack."""
        return self.get_pipette().get("return_tip_height", 0.5)

    def get_well_bottom_clearance(self) -> Clearances:
        """The distance above the bottom of a well to aspirate or dispense."""
        return self._well_bottom_clearances

    def get_flow_rate(self) -> FlowRates:
        return self._flow_rates

    def get_speed(self) -> PlungerSpeeds:
        return self._speeds

    def set_flow_rate(
        self,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        """Set the flow rates."""
        self._protocol_interface.get_hardware().hardware.set_flow_rate(
            mount=self._mount,
            aspirate=aspirate,
            dispense=dispense,
            blow_out=blow_out,
        )

    def set_pipette_speed(
        self,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        """Set pipette speeds."""
        self._protocol_interface.get_hardware().hardware.set_pipette_speed(
            mount=self._mount,
            aspirate=aspirate,
            dispense=dispense,
            blow_out=blow_out,
        )

    def pair_with(
        self, other_instrument: AbstractInstrument
    ) -> AbstractPairedInstrument:
        """Create an AbstractPairedInstrument"""
        return PairedInstrument(
            primary_instrument=self,
            secondary_instrument=other_instrument,
            ctx=self._protocol_interface,
            pair_policy=PipettePair.of_mount(self.get_mount()),
            hardware_manager=self._protocol_interface.get_hardware(),
        )



"""The interface that implements InstrumentContext."""

from __future__ import annotations

from abc import abstractmethod, ABC
import typing

from opentrons import types
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocols.api_support.util import Clearances, PlungerSpeeds, FlowRates
from opentrons.protocols.context.paired_instrument import AbstractPairedInstrument
from opentrons.protocols.context.well import WellImplementation


class AbstractInstrument(ABC):
    @abstractmethod
    def get_default_speed(self) -> float:
        ...

    @abstractmethod
    def set_default_speed(self, speed: float) -> None:
        ...

    @abstractmethod
    def aspirate(self, volume: float, rate: float) -> None:
        ...

    @abstractmethod
    def dispense(self, volume: float, rate: float) -> None:
        ...

    @abstractmethod
    def blow_out(self) -> None:
        ...

    @abstractmethod
    def touch_tip(
        self, location: WellImplementation, radius: float, v_offset: float, speed: float
    ) -> None:
        ...

    @abstractmethod
    def pick_up_tip(
        self,
        well: WellImplementation,
        tip_length: float,
        presses: typing.Optional[int],
        increment: typing.Optional[float],
    ) -> None:
        ...

    @abstractmethod
    def drop_tip(self, home_after: bool) -> None:
        ...

    @abstractmethod
    def home(self) -> None:
        ...

    @abstractmethod
    def home_plunger(self) -> None:
        ...

    @abstractmethod
    def delay(self) -> None:
        ...

    @abstractmethod
    def move_to(
        self,
        location: types.Location,
        force_direct: bool,
        minimum_z_height: typing.Optional[float],
        speed: typing.Optional[float],
    ) -> None:
        ...

    @abstractmethod
    def get_mount(self) -> types.Mount:
        ...

    @abstractmethod
    def get_instrument_name(self) -> str:
        ...

    @abstractmethod
    def get_pipette_name(self) -> str:
        ...

    @abstractmethod
    def get_model(self) -> str:
        ...

    @abstractmethod
    def get_min_volume(self) -> float:
        ...

    @abstractmethod
    def get_max_volume(self) -> float:
        ...

    @abstractmethod
    def get_current_volume(self) -> float:
        ...

    @abstractmethod
    def get_available_volume(self) -> float:
        ...

    @abstractmethod
    def get_pipette(self) -> PipetteDict:
        ...

    @abstractmethod
    def get_channels(self) -> int:
        ...

    @abstractmethod
    def has_tip(self) -> bool:
        ...

    @abstractmethod
    def is_ready_to_aspirate(self) -> bool:
        ...

    @abstractmethod
    def prepare_for_aspirate(self) -> None:
        ...

    @abstractmethod
    def get_return_height(self) -> float:
        ...

    @abstractmethod
    def get_well_bottom_clearance(self) -> Clearances:
        ...

    @abstractmethod
    def get_speed(self) -> PlungerSpeeds:
        ...

    @abstractmethod
    def get_flow_rate(self) -> FlowRates:
        ...

    @abstractmethod
    def set_flow_rate(
        self,
        aspirate: typing.Optional[float] = None,
        dispense: typing.Optional[float] = None,
        blow_out: typing.Optional[float] = None,
    ) -> None:
        ...

    @abstractmethod
    def set_pipette_speed(
        self,
        aspirate: typing.Optional[float] = None,
        dispense: typing.Optional[float] = None,
        blow_out: typing.Optional[float] = None,
    ) -> None:
        ...

    @abstractmethod
    def pair_with(
        self, other_instrument: AbstractInstrument
    ) -> AbstractPairedInstrument:
        ...



import typing
from datetime import datetime, timezone


def deep_get(
    obj: typing.Union[typing.Mapping, typing.Sequence],
    key: typing.Sequence[typing.Union[str, int]],
    default=None,
):
    """
    Utility to get deeply nested element in a list, tuple or dict without
     resorting to some_dict.get('k1', {}).get('k2', {}).get('k3', {})....etc.

    :param obj: A dict, list, or tuple
    :param key: collection of keys
    :param default: the default to return on error
    :return: value or default
    """
    if not key:
        return default

    for k in key:
        try:
            obj = obj[k]  # type: ignore
        except (KeyError, TypeError, IndexError):
            return default

    return obj


def utc_now() -> datetime:
    """Return the UTC time with timezone"""
    return datetime.now(tz=timezone.utc)



import logging
import asyncio
import os
import re
from typing import List, Tuple, Optional
from glob import glob

from opentrons.config import IS_ROBOT, IS_LINUX
from opentrons.drivers.rpi_drivers import types
from opentrons.hardware_control.modules import ModuleAtPort

from .execution_manager import ExecutionManager
from .types import AionotifyEvent
from . import modules


log = logging.getLogger(__name__)

MODULE_PORT_REGEX = re.compile("|".join(modules.MODULE_HW_BY_NAME.keys()), re.I)


class AttachedModulesControl:
    """
    A class to handle monitoring module attachment, capturing the physical
    USB port information and finally building a module object.
    """

    def __init__(self, api):
        self._available_modules: List[modules.AbstractModule] = []
        self._api = api

    @classmethod
    async def build(cls, api_instance):
        mc_instance = cls(api_instance)
        if not api_instance.is_simulator:
            await mc_instance.register_modules(mc_instance.scan())
        return mc_instance

    @property
    def available_modules(self) -> List[modules.AbstractModule]:
        return self._available_modules

    @property
    def api(self):
        return self._api

    async def build_module(
        self,
        port: str,
        usb_port: types.USBPort,
        model: str,
        loop: asyncio.AbstractEventLoop,
        sim_model: str = None,
    ) -> modules.AbstractModule:
        return await modules.build(
            port=port,
            usb_port=usb_port,
            which=model,
            simulating=self.api.is_simulator,
            interrupt_callback=self.api.pause_with_message,
            loop=loop,
            execution_manager=self.api._execution_manager,
            sim_model=sim_model,
        )

    async def unregister_modules(
        self, mods_at_ports: List[modules.ModuleAtPort]
    ) -> None:
        """
        De-register Modules.

        Remove any modules that are no longer found by aionotify.
        """
        removed_modules = []
        for mod in mods_at_ports:
            for attached_mod in self.available_modules:
                if attached_mod.port == mod.port:
                    removed_modules.append(attached_mod)
        for removed_mod in removed_modules:
            try:
                self._available_modules.remove(removed_mod)
            except ValueError:
                log.exception(
                    f"Removed Module {removed_mod} not" " found in attached modules"
                )
        for removed_mod in removed_modules:
            log.info(
                f"Module {removed_mod.name()} detached" f" from port {removed_mod.port}"
            )
            await removed_mod.cleanup()

    async def register_modules(
        self,
        new_mods_at_ports: List[modules.ModuleAtPort] = None,
        removed_mods_at_ports: List[modules.ModuleAtPort] = None,
    ) -> None:
        """
        Register Modules.

        Upon system recognition of a module being plugged in, we should
        register that module and de-register any modules that are
        no longer found on the system.
        """
        if new_mods_at_ports is None:
            new_mods_at_ports = []
        if removed_mods_at_ports is None:
            removed_mods_at_ports = []

        # destroy removed mods
        await self.unregister_modules(removed_mods_at_ports)
        sorted_mods_at_port = self.api._backend._usb.match_virtual_ports(
            new_mods_at_ports
        )

        # build new mods
        for mod in sorted_mods_at_port:
            new_instance = await self.build_module(
                port=mod.port, usb_port=mod.usb_port, model=mod.name, loop=self.api.loop
            )
            self._available_modules.append(new_instance)
            log.info(
                f"Module {mod.name} discovered and attached"
                f" at port {mod.port}, new_instance: {new_instance}"
            )

    async def parse_modules(
        self,
        by_model: modules.types.ModuleModel,
        resolved_type: modules.types.ModuleType,
    ) -> Tuple[List[modules.AbstractModule], Optional[modules.AbstractModule]]:
        """
        Parse Modules.
        Given a module model and type, find all attached
        modules that fit this criteria. If there are no
        modules attached, but the module is being loaded
        in simulation, then it should return a simulating
        module of the same type.
        """
        matching_modules = []
        simulated_module = None
        mod_type = {
            modules.types.ModuleType.MAGNETIC: "magdeck",
            modules.types.ModuleType.TEMPERATURE: "tempdeck",
            modules.types.ModuleType.THERMOCYCLER: "thermocycler",
        }[resolved_type]
        for module in self.available_modules:
            if mod_type == module.name():
                matching_modules.append(module)
        if self.api.is_simulator:
            module_builder = {
                "magdeck": modules.MagDeck.build,
                "tempdeck": modules.TempDeck.build,
                "thermocycler": modules.Thermocycler.build,
            }[mod_type]
            if module_builder:
                simulating_module = await module_builder(
                    port="",
                    usb_port=self.api._backend._usb.find_port(""),
                    simulating=True,
                    loop=self.api.loop,
                    execution_manager=ExecutionManager(loop=self.api.loop),
                    sim_model=by_model.value,
                )
                simulated_module = simulating_module
        return matching_modules, simulated_module

    def scan(self) -> List[modules.ModuleAtPort]:
        """Scan for connected modules and return list of
        tuples of serial ports and device names
        """
        if IS_ROBOT and IS_LINUX:
            devices = glob("/dev/ot_module*")
        else:
            devices = []

        discovered_modules = []

        for port in devices:
            symlink_port = port.split("dev/")[1]
            module_at_port = self.get_module_at_port(symlink_port)
            if module_at_port:
                discovered_modules.append(module_at_port)

        # Check for emulator environment variables
        emulator_uri = os.environ.get("OT_THERMOCYCLER_EMULATOR_URI")
        if emulator_uri:
            discovered_modules.append(
                ModuleAtPort(port=emulator_uri, name="thermocycler")
            )

        emulator_uri = os.environ.get("OT_TEMPERATURE_EMULATOR_URI")
        if emulator_uri:
            discovered_modules.append(ModuleAtPort(port=emulator_uri, name="tempdeck"))

        emulator_uri = os.environ.get("OT_MAGNETIC_EMULATOR_URI")
        if emulator_uri:
            discovered_modules.append(ModuleAtPort(port=emulator_uri, name="magdeck"))

        log.debug("Discovered modules: {}".format(discovered_modules))
        return discovered_modules

    @staticmethod
    def get_module_at_port(port: str) -> Optional[modules.ModuleAtPort]:
        """Given a port, returns either a ModuleAtPort
        if it is a recognized module, or None if not recognized.
        """
        match = MODULE_PORT_REGEX.search(port)
        if match:
            name = match.group().lower()
            if name not in modules.MODULE_HW_BY_NAME:
                log.warning(f"Unexpected module connected: {name} on {port}")
                return None
            return modules.ModuleAtPort(port=f"/dev/{port}", name=name)
        return None

    async def handle_module_appearance(self, event: AionotifyEvent):
        """Only called upon availability of aionotify. Check that
        the file system has changed and either remove or add modules
        depending on the result.

        :param event_name: The title of the even passed into aionotify.
        :param event_flags: AionotifyFlags dataclass that maps flags listed from
        the aionotify event.
        """
        maybe_module_at_port = self.get_module_at_port(event.name)
        new_modules = None
        removed_modules = None
        if maybe_module_at_port is not None:
            if hasattr(event.flags, "DELETE"):
                removed_modules = [maybe_module_at_port]
                log.info(f"Module Removed: {maybe_module_at_port}")
            elif hasattr(event.flags, "CREATE"):
                new_modules = [maybe_module_at_port]
                log.info(f"Module Added: {maybe_module_at_port}")
            try:
                await self.register_modules(
                    removed_mods_at_ports=removed_modules,
                    new_mods_at_ports=new_modules,
                )
            except Exception:
                log.exception("Exception in Module registration")



from asyncio import AbstractEventLoop
from typing import Union
from opentrons.broker import Notifications, Broker
from opentrons.hardware_control import ThreadManager, ThreadedAsyncLock
from .session import SessionManager, Session
from .dev_types import Message as SessionMessage
from .calibration import CalibrationManager, Message as CalibrationMessage


class MainRouter:
    def __init__(
        self,
        hardware: ThreadManager,
        loop: AbstractEventLoop = None,
        lock: ThreadedAsyncLock = None,
    ) -> None:
        topics = [Session.TOPIC, CalibrationManager.TOPIC]
        self._broker = Broker()
        self._notifications: Notifications[
            Union[SessionMessage, CalibrationMessage]
        ] = Notifications(topics, self._broker, loop=loop)

        checked_hw = hardware.sync
        self.session_manager = SessionManager(
            hardware=checked_hw, loop=loop, broker=self._broker, lock=lock
        )
        self.calibration_manager = CalibrationManager(
            hardware=checked_hw, loop=loop, broker=self._broker, lock=lock
        )

    @property
    def notifications(self) -> Notifications[Union[SessionMessage, CalibrationMessage]]:
        return self._notifications

    @property
    def broker(self) -> Broker:
        return self._broker



"""Aspirate command request, result, and implementation models."""
from __future__ import annotations
from typing import Optional, Type
from typing_extensions import Literal


from .pipetting_common import BaseLiquidHandlingData, BaseLiquidHandlingResult
from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest

AspirateCommandType = Literal["aspirate"]


class AspirateData(BaseLiquidHandlingData):
    """Data required to aspirate from a specific well."""

    pass


class AspirateResult(BaseLiquidHandlingResult):
    """Result data from the execution of a AspirateRequest."""

    pass


class AspirateImplementation(AbstractCommandImpl[AspirateData, AspirateResult]):
    """Aspirate command implementation."""

    async def execute(self, data: AspirateData) -> AspirateResult:
        """Move to and aspirate from the requested well."""
        volume = await self._pipetting.aspirate(
            pipette_id=data.pipetteId,
            labware_id=data.labwareId,
            well_name=data.wellName,
            well_location=data.wellLocation,
            volume=data.volume,
        )

        return AspirateResult(volume=volume)


class Aspirate(BaseCommand[AspirateData, AspirateResult]):
    """Aspirate command model."""

    commandType: AspirateCommandType = "aspirate"
    data: AspirateData
    result: Optional[AspirateResult]

    _ImplementationCls: Type[AspirateImplementation] = AspirateImplementation


class AspirateRequest(BaseCommandRequest[AspirateData]):
    """Create aspirate command request model."""

    commandType: AspirateCommandType = "aspirate"
    data: AspirateData

    _CommandCls: Type[Aspirate] = Aspirate



"""State store actions.

Actions can be passed to the StateStore, where they will trigger
state changes.
"""

from dataclasses import dataclass
from typing import Union

from ..commands import Command


@dataclass(frozen=True)
class PlayAction:
    """Start or resume processing commands in the engine."""

    pass


@dataclass(frozen=True)
class PauseAction:
    """Pause processing commands in the engine."""

    pass


@dataclass(frozen=True)
class StopAction:
    """Stop processing commands in the engine, marking the engine status as done."""

    pass


@dataclass(frozen=True)
class UpdateCommandAction:
    """Update a given command."""

    command: Command


Action = Union[
    PlayAction,
    PauseAction,
    StopAction,
    UpdateCommandAction,
]



"""Python Protocol API v3 errors."""
# TODO(mc, 2021-04-22): assign unique codes to all these errors
# TODO(mc, 2021-04-22): explore how ProtocolEngine and ProtocolAPI errors
# interact and combine
from opentrons.protocol_engine.errors import LabwareIsNotTipRackError


class InvalidPipetteNameError(ValueError):
    """Error raised if an invalid pipette name is used."""

    def __init__(self, invalid_value: str) -> None:
        """Initialize the error and message with the invalid value."""
        super().__init__(f"{invalid_value} is not a valid pipette name.")
        self.invalid_value = invalid_value


class InvalidMountError(ValueError):
    """Error raised if an invalid mount is used."""

    def __init__(self, invalid_value: str) -> None:
        """Initialize the error and message with the invalid value."""
        super().__init__(f"{invalid_value} is not a valid mount.")
        self.invalid_value = invalid_value


__all__ = [
    # re-exports from opentrons.protocol_engine
    "LabwareIsNotTipRackError",
    # exports specific to the Protocol API layer
    "InvalidPipetteNameError",
    "InvalidMountError",
]



from .mod_abc import AbstractModule
from .tempdeck import TempDeck
from .magdeck import MagDeck
from .thermocycler import Thermocycler
from .update import update_firmware
from .utils import MODULE_HW_BY_NAME, build
from .types import (
    ThermocyclerStep,
    InterruptCallback,
    UploadFunction,
    BundledFirmware,
    UpdateError,
    ModuleAtPort,
)

__all__ = [
    "MODULE_HW_BY_NAME",
    "build",
    "update_firmware",
    "ThermocyclerStep",
    "AbstractModule",
    "TempDeck",
    "MagDeck",
    "Thermocycler",
    "InterruptCallback",
    "UploadFunction",
    "BundledFirmware",
    "UpdateError",
    "ModuleAtPort",
]



class ExceptionInProtocolError(Exception):
    """This exception wraps an exception that was raised from a protocol
    for proper error message formatting by the rpc, since it's only here that
    we can properly figure out formatting
    """

    def __init__(self, original_exc, original_tb, message, line):
        self.original_exc = original_exc
        self.original_tb = original_tb
        self.message = message
        self.line = line
        super().__init__(original_exc, original_tb, message, line)

    def __str__(self):
        return "{}{}: {}".format(
            self.original_exc.__class__.__name__,
            " [line {}]".format(self.line) if self.line else "",
            self.message,
        )



"""Pause protocol command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import Optional, Type
from typing_extensions import Literal

from .command import AbstractCommandImpl, BaseCommand, BaseCommandRequest


PauseCommandType = Literal["pause"]


class PauseData(BaseModel):
    """Data required to pause the protocol."""

    message: Optional[str] = Field(
        None,
        description="A user-facing message associated with the pause",
    )


class PauseResult(BaseModel):
    """Result data from the execution of a Pause command."""

    pass


class PauseImplementation(AbstractCommandImpl[PauseData, PauseResult]):
    """Pause command implementation."""

    async def execute(self, data: PauseData) -> PauseResult:
        """Dispatch a PauseAction to the store to pause the protocol."""
        await self._run_control.pause()
        return PauseResult()


class Pause(BaseCommand[PauseData, PauseResult]):
    """Pause command model."""

    commandType: PauseCommandType = "pause"
    data: PauseData
    result: Optional[PauseResult]

    _ImplementationCls: Type[PauseImplementation] = PauseImplementation


class PauseRequest(BaseCommandRequest[PauseData]):
    """Pause command request model."""

    commandType: PauseCommandType = "pause"
    data: PauseData

    _CommandCls: Type[Pause] = Pause



import asyncio
import logging
import os
from pathlib import Path
from glob import glob
from typing import Any, Dict, Tuple, Optional
from .types import UpdateError
from .mod_abc import AbstractModule

log = logging.getLogger(__name__)


async def update_firmware(
    module: AbstractModule,
    firmware_file: str,
    loop: Optional[asyncio.AbstractEventLoop],
):
    """Apply update of given firmware file to given module.

    raises an UpdateError with the reason for the failure.
    """
    flash_port = await module.prep_for_update()
    kwargs: Dict[str, Any] = {
        "stdout": asyncio.subprocess.PIPE,
        "stderr": asyncio.subprocess.PIPE,
        "loop": loop,
    }
    successful, res = await module.bootloader()(flash_port, firmware_file, kwargs)
    if not successful:
        log.info(f"Bootloader reponse: {res}")
        raise UpdateError(res)


async def find_bootloader_port():
    """
    Finds the port of an Opentrons Module that has entered its bootloader.
    The bootloader port shows up as 'ot_module_(avrdude|samba)_bootloader'
    on the pi; return found port.
    """

    for attempt in range(3):
        bootloader_ports = glob("/dev/ot_module_*_bootloader*")
        if bootloader_ports:
            if len(bootloader_ports) == 1:
                log.info(f"Found bootloader at port {bootloader_ports[0]}")
                return bootloader_ports[0]
            elif len(bootloader_ports) > 1:
                raise OSError("Multiple new bootloader ports" "found on mode switch")
        await asyncio.sleep(2)
    raise Exception("No ot_module bootloaders found in /dev. Try again")


async def upload_via_avrdude(
    port: str, firmware_file_path: str, kwargs: Dict[str, Any]
) -> Tuple[bool, str]:
    """
    Run firmware upload command for hardware module with avrdude bootloader.

    Returns tuple of success boolean and message from bootloader.
    """
    # avrdude_options
    PART_NO = "atmega32u4"
    PROGRAMMER_ID = "avr109"
    BAUDRATE = "57600"

    config_file_path = Path("/etc/avrdude.conf")
    proc = await asyncio.create_subprocess_exec(
        "avrdude",
        "-C{}".format(config_file_path),
        "-v",
        "-p{}".format(PART_NO),
        "-c{}".format(PROGRAMMER_ID),
        "-P{}".format(port),
        "-b{}".format(BAUDRATE),
        "-D",
        "-Uflash:w:{}:i".format(firmware_file_path),
        **kwargs,
    )
    await proc.wait()

    _result = await proc.communicate()
    result = _result[1].decode()
    avrdude_res = _format_avrdude_response(result)
    if avrdude_res[0]:
        log.debug(result)
    else:
        log.error(
            "Failed to update module firmware for {}: {}".format(port, avrdude_res[1])
        )
    return avrdude_res


def _format_avrdude_response(raw_response: str) -> Tuple[bool, str]:
    avrdude_log = ""
    for line in raw_response.splitlines():
        if "avrdude:" in line and line != raw_response.splitlines()[1]:
            avrdude_log += line.lstrip("avrdude:") + ".."
            if "flash verified" in line:
                return True, line.lstrip("avrdude: ")
    return False, avrdude_log


async def upload_via_bossa(
    port: str, firmware_file_path: str, kwargs: Dict[str, Any]
) -> Tuple[bool, str]:
    """
    Run firmware upload command for hardware module with SAMBA bootloader.

    Returns tuple of success boolean and message from bootloader.
    """
    # bossac -p/dev/ttyACM1 -e -w -v -R --offset=0x2000
    #   modules/thermo-cycler/production/firmware/thermo-cycler-arduino.ino.bin
    # NOTE: bossac cannot traverse symlinks to port,
    # so we resolve to real path
    resolved_symlink = os.path.realpath(port)
    log.info(
        f"device at symlinked port: {port} " f"resolved to path: {resolved_symlink}"
    )
    bossa_args = [
        "bossac",
        f"-p{resolved_symlink}",
        "-e",
        "-w",
        "-v",
        "-R",
        "--offset=0x2000",
        f"{firmware_file_path}",
    ]

    proc = await asyncio.create_subprocess_exec(*bossa_args, **kwargs)
    stdout, stderr = await proc.communicate()
    res = stdout.decode()
    if "Verify successful" in res:
        log.debug(res)
        return True, res
    elif stderr:
        log.error(f"Failed to update module firmware for {port}: {res}")
        log.error(f"Error given: {stderr.decode()}")
        return False, res
    return False, ""



from .serial_connection import SerialConnection
from opentrons.drivers.asyncio.communication.errors import (
    SerialException,
    NoResponse,
    AlarmResponse,
    ErrorResponse,
)
from .async_serial import AsyncSerial

__all__ = [
    "SerialConnection",
    "AsyncSerial",
    "SerialException",
    "NoResponse",
    "AlarmResponse",
    "ErrorResponse",
]
